@@ -1,9 +1,13 @@
+# Copyright Contributors to the Amundsen project.
+# SPDX-License-Identifier: Apache-2.0
+
 from abc import ABCMeta, abstractmethod
-from typing import Any, Dict, List, Union
+from typing import Any, Dict, List, Union, Optional
 
 from amundsen_common.models.popular_table import PopularTable
 from amundsen_common.models.table import Table
-from amundsen_common.models.user import User as UserEntity
+from amundsen_common.models.user import User
+from amundsen_common.models.dashboard import DashboardSummary
 
 from metadata_service.entity.dashboard_detail import DashboardDetail as DashboardDetailEntity
 from metadata_service.entity.description import Description
@@ -18,11 +22,11 @@ class BaseProxy(metaclass=ABCMeta):
     """
 
     @abstractmethod
-    def get_user(self, *, id: str) -> Union[UserEntity, None]:
+    def get_user(self, *, id: str) -> Union[User, None]:
         pass
 
     @abstractmethod
-    def get_users(self) -> List[UserEntity]:
+    def get_users(self) -> List[User]:
         pass
 
     @abstractmethod
@@ -52,10 +56,20 @@ class BaseProxy(metaclass=ABCMeta):
     def add_tag(self, *, id: str, tag: str, tag_type: str, resource_type: ResourceType) -> None:
         pass
 
+    @abstractmethod
+    def add_badge(self, *, id: str, badge_name: str, category: str = '',
+                  resource_type: ResourceType) -> None:
+        pass
+
     @abstractmethod
     def delete_tag(self, *, id: str, tag: str, tag_type: str, resource_type: ResourceType) -> None:
         pass
 
+    @abstractmethod
+    def delete_badge(self, *, id: str, badge_name: str, category: str,
+                     resource_type: ResourceType) -> None:
+        pass
+
     @abstractmethod
     def put_column_description(self, *,
                                table_uri: str,
@@ -70,7 +84,9 @@ class BaseProxy(metaclass=ABCMeta):
         pass
 
     @abstractmethod
-    def get_popular_tables(self, *, num_entries: int) -> List[PopularTable]:
+    def get_popular_tables(self, *,
+                           num_entries: int,
+                           user_id: Optional[str] = None) -> List[PopularTable]:
         pass
 
     @abstractmethod
@@ -81,6 +97,15 @@ class BaseProxy(metaclass=ABCMeta):
     def get_tags(self) -> List:
         pass
 
+    @abstractmethod
+    def get_badges(self) -> List:
+        pass
+
+    @abstractmethod
+    def get_dashboard_by_user_relation(self, *, user_email: str, relation_type: UserResourceRel) \
+            -> Dict[str, List[DashboardSummary]]:
+        pass
+
     @abstractmethod
     def get_table_by_user_relation(self, *, user_email: str,
                                    relation_type: UserResourceRel) -> Dict[str, Any]:
@@ -122,3 +147,9 @@ class BaseProxy(metaclass=ABCMeta):
                                   id: str,
                                   description: str) -> None:
         pass
+
+    @abstractmethod
+    def get_resources_using_table(self, *,
+                                  id: str,
+                                  resource_type: ResourceType) -> Dict[str, List[DashboardSummary]]:
+        pass
