@@ -1,3 +1,6 @@
+# Copyright Contributors to the Amundsen project.
+# SPDX-License-Identifier: Apache-2.0
+
 import logging
 import textwrap
 import time
@@ -5,21 +8,26 @@ from random import randint
 from typing import (Any, Dict, List, Optional, Tuple, Union,  # noqa: F401
                     no_type_check)
 
+import neo4j
 from amundsen_common.models.dashboard import DashboardSummary
 from amundsen_common.models.popular_table import PopularTable
 from amundsen_common.models.table import (Application, Column, Reader, Source,
-                                          Statistics, Table, User,
-                                          Watermark, ProgrammaticDescription)
-from amundsen_common.models.table import Tag
+                                          Stat, Table, User,
+                                          Watermark, ProgrammaticDescription, Tag,
+                                          Badge as TableBadge)
 from amundsen_common.models.user import User as UserEntity
 from beaker.cache import CacheManager
 from beaker.util import parse_cache_config_options
-from neo4j.v1 import BoltStatementResult, Driver, GraphDatabase  # noqa: F401
+from flask import current_app, has_app_context
+from neo4j import BoltStatementResult, Driver, GraphDatabase  # noqa: F401
 
+from metadata_service import config
 from metadata_service.entity.dashboard_detail import DashboardDetail as DashboardDetailEntity
+from metadata_service.entity.dashboard_query import DashboardQuery as DashboardQueryEntity
 from metadata_service.entity.description import Description
 from metadata_service.entity.resource_type import ResourceType
 from metadata_service.entity.tag_detail import TagDetail
+from metadata_service.entity.badge import Badge
 from metadata_service.exception import NotFoundException
 from metadata_service.proxy.base_proxy import BaseProxy
 from metadata_service.proxy.statsd_utilities import timer_with_counter
@@ -44,7 +52,9 @@ class Neo4jProxy(BaseProxy):
                  user: str = 'neo4j',
                  password: str = '',
                  num_conns: int = 50,
-                 max_connection_lifetime_sec: int = 100) -> None:
+                 max_connection_lifetime_sec: int = 100,
+                 encrypted: bool = False,
+                 validate_ssl: bool = False) -> None:
         """
         There's currently no request timeout from client side where server
         side can be enforced via "dbms.transaction.timeout"
@@ -56,10 +66,22 @@ class Neo4jProxy(BaseProxy):
         value needs to be smaller than surrounding network environment's timeout.
         """
         endpoint = f'{host}:{port}'
+        LOGGER.info('NEO4J endpoint: {}'.format(endpoint))
+        trust = neo4j.TRUST_SYSTEM_CA_SIGNED_CERTIFICATES if validate_ssl else neo4j.TRUST_ALL_CERTIFICATES
         self._driver = GraphDatabase.driver(endpoint, max_connection_pool_size=num_conns,
                                             connection_timeout=10,
                                             max_connection_lifetime=max_connection_lifetime_sec,
-                                            auth=(user, password))  # type: Driver
+                                            auth=(user, password),
+                                            encrypted=encrypted,
+                                            trust=trust)  # type: Driver
+
+    def is_healthy(self) -> None:
+        # throws if cluster unhealthy or can't connect.  An alternative would be to use one of
+        # the HTTP status endpoints, which might be more specific, but don't implicitly test
+        # our configuration.
+        with self._driver.session() as session:
+            session.read_transaction(self._execute_cypher_query,
+                                     statement='CALL dbms.cluster.overview()', param_dict={})
 
     @timer_with_counter
     def get_table(self, *, table_uri: str) -> Table:
@@ -105,7 +127,9 @@ class Neo4jProxy(BaseProxy):
         OPTIONAL MATCH (tbl)-[:DESCRIPTION]->(tbl_dscrpt:Description)
         OPTIONAL MATCH (col:Column)-[:DESCRIPTION]->(col_dscrpt:Description)
         OPTIONAL MATCH (col:Column)-[:STAT]->(stat:Stat)
-        RETURN db, clstr, schema, tbl, tbl_dscrpt, col, col_dscrpt, collect(distinct stat) as col_stats
+        OPTIONAL MATCH (col:Column)-[:HAS_BADGE]->(badge:Badge)
+        RETURN db, clstr, schema, tbl, tbl_dscrpt, col, col_dscrpt, collect(distinct stat) as col_stats,
+        collect(distinct badge) as col_badges
         ORDER BY col.sort_order;""")
 
         tbl_col_neo4j_records = self._execute_cypher_query(
@@ -116,7 +140,7 @@ class Neo4jProxy(BaseProxy):
             # Getting last record from this for loop as Neo4j's result's random access is O(n) operation.
             col_stats = []
             for stat in tbl_col_neo4j_record['col_stats']:
-                col_stat = Statistics(
+                col_stat = Stat(
                     stat_type=stat['stat_name'],
                     stat_val=stat['stat_val'],
                     start_epoch=int(float(stat['start_epoch'])),
@@ -124,12 +148,17 @@ class Neo4jProxy(BaseProxy):
                 )
                 col_stats.append(col_stat)
 
+            column_badges = []
+            for badge in tbl_col_neo4j_record['col_badges']:
+                column_badges.append(TableBadge(badge_name=badge['key'], category=badge['category']))
+
             last_neo4j_record = tbl_col_neo4j_record
             col = Column(name=tbl_col_neo4j_record['col']['name'],
                          description=self._safe_get(tbl_col_neo4j_record, 'col_dscrpt', 'description'),
                          col_type=tbl_col_neo4j_record['col']['type'],
                          sort_order=int(tbl_col_neo4j_record['col']['sort_order']),
-                         stats=col_stats)
+                         stats=col_stats,
+                         badges=column_badges)
 
             cols.append(col)
 
@@ -174,7 +203,7 @@ class Neo4jProxy(BaseProxy):
         OPTIONAL MATCH (tbl)-[:LAST_UPDATED_AT]->(t:Timestamp)
         OPTIONAL MATCH (owner:User)<-[:OWNER]-(tbl)
         OPTIONAL MATCH (tbl)-[:TAGGED_BY]->(tag:Tag{tag_type: $tag_normal_type})
-        OPTIONAL MATCH (tbl)-[:TAGGED_BY]->(badge:Tag{tag_type: $tag_badge_type})
+        OPTIONAL MATCH (tbl)-[:HAS_BADGE]->(badge:Badge)
         OPTIONAL MATCH (tbl)-[:SOURCE]->(src:Source)
         OPTIONAL MATCH (tbl)-[:DESCRIPTION]->(prog_descriptions:Programmatic_Description)
         RETURN collect(distinct wmk) as wmk_records,
@@ -189,8 +218,7 @@ class Neo4jProxy(BaseProxy):
 
         table_records = self._execute_cypher_query(statement=table_level_query,
                                                    param_dict={'tbl_key': table_uri,
-                                                               'tag_normal_type': 'default',
-                                                               'tag_badge_type': 'badge'})
+                                                               'tag_normal_type': 'default'})
 
         table_records = table_records.single()
 
@@ -217,11 +245,12 @@ class Neo4jProxy(BaseProxy):
                 tags.append(tag_result)
 
         badges = []
+        # this is for any badges added with BadgeAPI instead of TagAPI
         if table_records.get('badge_records'):
             badge_records = table_records['badge_records']
             for record in badge_records:
-                badge_result = Tag(tag_name=record['key'],
-                                   tag_type=record['tag_type'])
+                badge_result = TableBadge(badge_name=record['key'],
+                                          category=record['category'])
                 badges.append(badge_result)
 
         application_record = table_records['application']
@@ -380,7 +409,6 @@ class Neo4jProxy(BaseProxy):
             raise e
 
         finally:
-            tx.close()
             if LOGGER.isEnabledFor(logging.DEBUG):
                 LOGGER.debug('Update process elapsed for {} seconds'.format(time.time() - start))
 
@@ -481,9 +509,6 @@ class Neo4jProxy(BaseProxy):
             raise e
 
         finally:
-
-            tx.close()
-
             if LOGGER.isEnabledFor(logging.DEBUG):
                 LOGGER.debug('Update process elapsed for {} seconds'.format(time.time() - start))
 
@@ -507,7 +532,7 @@ class Neo4jProxy(BaseProxy):
 
         upsert_owner_relation_query = textwrap.dedent("""
         MATCH (n1:User {key: $user_email}), (n2:Table {key: $tbl_key})
-        MERGE (n2)-[r2:OWNER]->(n1)
+        MERGE (n1)-[r1:OWNER_OF]->(n2)-[r2:OWNER]->(n1)
         RETURN n1.key, n2.key
         """)
 
@@ -522,14 +547,12 @@ class Neo4jProxy(BaseProxy):
                 raise RuntimeError('Failed to create relation between '
                                    'owner {owner} and table {tbl}'.format(owner=owner,
                                                                           tbl=table_uri))
+            tx.commit()
         except Exception as e:
             if not tx.closed():
                 tx.rollback()
             # propagate the exception back to api
             raise e
-        finally:
-            tx.commit()
-            tx.close()
 
     @timer_with_counter
     def delete_owner(self, *,
@@ -537,15 +560,16 @@ class Neo4jProxy(BaseProxy):
                      owner: str) -> None:
         """
         Delete the owner / owned_by relationship.
-
         :param table_uri:
         :param owner:
         :return:
         """
         delete_query = textwrap.dedent("""
-        MATCH (n1:User{key: $user_email})<-[r1:OWNER]-(n2:Table {key: $tbl_key}) DELETE r1
+        MATCH (n1:User{key: $user_email}), (n2:Table {key: $tbl_key})
+        OPTIONAL MATCH (n1)-[r1:OWNER_OF]->(n2)
+        OPTIONAL MATCH (n2)-[r2:OWNER]->(n1)
+        DELETE r1,r2
         """)
-
         try:
             tx = self._driver.session().begin_transaction()
             tx.run(delete_query, {'user_email': owner,
@@ -557,7 +581,101 @@ class Neo4jProxy(BaseProxy):
             raise e
         finally:
             tx.commit()
-            tx.close()
+
+    @timer_with_counter
+    def add_badge(self, *,
+                  id: str,
+                  badge_name: str,
+                  category: str = '',
+                  resource_type: ResourceType = ResourceType.Table) -> None:
+
+        LOGGER.info('New badge {} for id {} with category {} '
+                    'and resource type {}'.format(badge_name, id, category, resource_type.name))
+
+        validation_query = \
+            'MATCH (n:{resource_type} {{key: $key}}) return n'.format(resource_type=resource_type.name)
+
+        upsert_badge_query = textwrap.dedent("""
+        MERGE (u:Badge {key: $badge_name})
+        on CREATE SET u={key: $badge_name, category: $category}
+        on MATCH SET u={key: $badge_name, category: $category}
+        """)
+
+        upsert_badge_relation_query = textwrap.dedent("""
+        MATCH(n1:Badge {{key: $badge_name, category: $category}}),
+        (n2:{resource_type} {{key: $key}})
+        MERGE (n1)-[r1:BADGE_FOR]->(n2)-[r2:HAS_BADGE]->(n1)
+        RETURN n1.key, n2.key
+        """.format(resource_type=resource_type.name))
+
+        try:
+            tx = self._driver.session().begin_transaction()
+            tbl_result = tx.run(validation_query, {'key': id})
+            if not tbl_result.single():
+                raise NotFoundException('id {} does not exist'.format(id))
+
+            tx.run(upsert_badge_query, {'badge_name': badge_name,
+                                        'category': category})
+
+            result = tx.run(upsert_badge_relation_query, {'badge_name': badge_name,
+                                                          'key': id,
+                                                          'category': category})
+
+            if not result.single():
+                raise RuntimeError('failed to create relation between '
+                                   'badge {badge} and resource {resource} of resource type '
+                                   '{resource_type} MORE {q}'.format(badge=badge_name,
+                                                                     resource=id,
+                                                                     resource_type=resource_type,
+                                                                     q=upsert_badge_relation_query))
+            tx.commit()
+        except Exception as e:
+            if not tx.closed():
+                tx.rollback()
+            raise e
+
+    @timer_with_counter
+    def delete_badge(self, id: str,
+                     badge_name: str,
+                     category: str,
+                     resource_type: ResourceType = ResourceType.Table) -> None:
+
+        # TODO for some reason when deleting it will say it was successful
+        # even when the badge never existed to begin with
+        LOGGER.info('Delete badge {} for id {} with category {}'.format(badge_name, id, category))
+
+        # only deletes relationshop between badge and resource
+        delete_query = textwrap.dedent("""
+        MATCH (b:Badge {{key:$badge_name, category:$category}})-
+        [r1:BADGE_FOR]->(n:{resource_type} {{key: $key}})-[r2:HAS_BADGE]->(b) DELETE r1,r2
+        """.format(resource_type=resource_type.name))
+
+        try:
+            tx = self._driver.session().begin_transaction()
+            tx.run(delete_query, {'badge_name': badge_name,
+                                  'key': id,
+                                  'category': category})
+            tx.commit()
+        except Exception as e:
+            # propagate the exception back to api
+            if not tx.closed():
+                tx.rollback()
+            raise e
+
+    @timer_with_counter
+    def get_badges(self) -> List:
+        LOGGER.info('Get all badges')
+        query = textwrap.dedent("""
+        MATCH (b:Badge) RETURN b as badge
+        """)
+        records = self._execute_cypher_query(statement=query,
+                                             param_dict={})
+        results = []
+        for record in records:
+            results.append(Badge(badge_name=record['badge']['key'],
+                                 category=record['badge']['category']))
+
+        return results
 
     @timer_with_counter
     def add_tag(self, *,
@@ -600,7 +718,7 @@ class Neo4jProxy(BaseProxy):
             if not tbl_result.single():
                 raise NotFoundException('id {} does not exist'.format(id))
 
-            # upsert the node. Currently the type for all the tags is default. We could change it later per UI.
+            # upsert the node
             tx.run(upsert_tag_query, {'tag': tag,
                                       'tag_type': tag_type})
             result = tx.run(upsert_tag_relation_query, {'tag': tag,
@@ -618,9 +736,6 @@ class Neo4jProxy(BaseProxy):
                 tx.rollback()
             # propagate the exception back to api
             raise e
-        finally:
-            if not tx.closed():
-                tx.close()
 
     @timer_with_counter
     def delete_tag(self, *,
@@ -652,14 +767,12 @@ class Neo4jProxy(BaseProxy):
             tx.run(delete_query, {'tag': tag,
                                   'key': id,
                                   'tag_type': tag_type})
+            tx.commit()
         except Exception as e:
             # propagate the exception back to api
             if not tx.closed():
                 tx.rollback()
             raise e
-        finally:
-            tx.commit()
-            tx.close()
 
     @timer_with_counter
     def get_tags(self) -> List:
@@ -672,8 +785,10 @@ class Neo4jProxy(BaseProxy):
         # todo: Currently all the tags are default type, we could open it up if we want to include badge
         query = textwrap.dedent("""
         MATCH (t:Tag{tag_type: 'default'})
-        OPTIONAL MATCH (tbl:Table)-[:TAGGED_BY]->(t)
-        RETURN t as tag_name, count(distinct tbl.key) as tag_count
+        OPTIONAL MATCH (resource)-[:TAGGED_BY]->(t)
+        WITH t as tag_name, count(distinct resource.key) as tag_count
+        WHERE tag_count > 0
+        RETURN tag_name, tag_count
         """)
 
         records = self._execute_cypher_query(statement=query,
@@ -704,8 +819,8 @@ class Neo4jProxy(BaseProxy):
             return None
 
     @timer_with_counter
-    @_CACHE.cache('_get_popular_tables_uris', _GET_POPULAR_TABLE_CACHE_EXPIRY_SEC)
-    def _get_popular_tables_uris(self, num_entries: int) -> List[str]:
+    @_CACHE.cache('_get_global_popular_tables_uris', expire=_GET_POPULAR_TABLE_CACHE_EXPIRY_SEC)
+    def _get_global_popular_tables_uris(self, num_entries: int) -> List[str]:
         """
         Retrieve popular table uris. Will provide tables with top x popularity score.
         Popularity score = number of distinct readers * log(total number of reads)
@@ -719,19 +834,54 @@ class Neo4jProxy(BaseProxy):
         query = textwrap.dedent("""
         MATCH (tbl:Table)-[r:READ_BY]->(u:User)
         WITH tbl.key as table_key, count(distinct u) as readers, sum(r.read_count) as total_reads
-        WHERE readers > 10
+        WHERE readers >= $num_readers
         RETURN table_key, readers, total_reads, (readers * log(total_reads)) as score
         ORDER BY score DESC LIMIT $num_entries;
         """)
-
         LOGGER.info('Querying popular tables URIs')
+        num_readers = current_app.config['POPULAR_TABLE_MINIMUM_READER_COUNT']
         records = self._execute_cypher_query(statement=query,
-                                             param_dict={'num_entries': num_entries})
+                                             param_dict={'num_readers': num_readers,
+                                                         'num_entries': num_entries})
 
         return [record['table_key'] for record in records]
 
     @timer_with_counter
-    def get_popular_tables(self, *, num_entries: int) -> List[PopularTable]:
+    @_CACHE.cache('_get_personal_popular_tables_uris', _GET_POPULAR_TABLE_CACHE_EXPIRY_SEC)
+    def _get_personal_popular_tables_uris(self, num_entries: int,
+                                          user_id: str) -> List[str]:
+        """
+        Retrieve personalized popular table uris. Will provide tables with top
+        popularity score that have been read by a peer of the user_id provided.
+        The popularity score is defined in the same way as `_get_global_popular_tables_uris`
+
+        The result of this method will be cached based on the key (num_entries, user_id),
+        and the cache will be expired based on _GET_POPULAR_TABLE_CACHE_EXPIRY_SEC
+
+        :return: Iterable of table uri
+        """
+        statement = textwrap.dedent("""
+        MATCH (:User {key:$user_id})<-[:READ_BY]-(:Table)-[:READ_BY]->
+             (coUser:User)<-[coRead:READ_BY]-(table:Table)
+        WITH table.key AS table_key, count(DISTINCT coUser) AS co_readers,
+             sum(coRead.read_count) AS total_co_reads
+        WHERE co_readers >= $num_readers
+        RETURN table_key, (co_readers * log(total_co_reads)) AS score
+        ORDER BY score DESC LIMIT $num_entries;
+        """)
+        LOGGER.info('Querying popular tables URIs')
+        num_readers = current_app.config['POPULAR_TABLE_MINIMUM_READER_COUNT']
+        records = self._execute_cypher_query(statement=statement,
+                                             param_dict={'user_id': user_id,
+                                                         'num_readers': num_readers,
+                                                         'num_entries': num_entries})
+
+        return [record['table_key'] for record in records]
+
+    @timer_with_counter
+    def get_popular_tables(self, *,
+                           num_entries: int,
+                           user_id: Optional[str] = None) -> List[PopularTable]:
         """
         Retrieve popular tables. As popular table computation requires full scan of table and user relationship,
         it will utilize cached method _get_popular_tables_uris.
@@ -739,8 +889,13 @@ class Neo4jProxy(BaseProxy):
         :param num_entries:
         :return: Iterable of PopularTable
         """
+        if user_id is None:
+            # Get global popular table URIs
+            table_uris = self._get_global_popular_tables_uris(num_entries)
+        else:
+            # Get personalized popular table URIs
+            table_uris = self._get_personal_popular_tables_uris(num_entries, user_id)
 
-        table_uris = self._get_popular_tables_uris(num_entries)
         if not table_uris:
             return []
 
@@ -810,6 +965,19 @@ class Neo4jProxy(BaseProxy):
 
     @staticmethod
     def _build_user_from_record(record: dict, manager_name: str = '') -> UserEntity:
+        """
+        Builds user record from Cypher query result. Other than the one defined in amundsen_common.models.user.User,
+        you could add more fields from User node into the User model by specifying keys in config.USER_OTHER_KEYS
+        :param record:
+        :param manager_name:
+        :return:
+        """
+        other_key_values = {}
+        if has_app_context() and current_app.config[config.USER_OTHER_KEYS]:
+            for k in current_app.config[config.USER_OTHER_KEYS]:
+                if k in record:
+                    other_key_values[k] = record[k]
+
         return UserEntity(email=record['email'],
                           first_name=record.get('first_name'),
                           last_name=record.get('last_name'),
@@ -820,7 +988,8 @@ class Neo4jProxy(BaseProxy):
                           slack_id=record.get('slack_id'),
                           employee_type=record.get('employee_type'),
                           role_name=record.get('role_name'),
-                          manager_fullname=record.get('manager_fullname', manager_name))
+                          manager_fullname=record.get('manager_fullname', manager_name),
+                          other_key_values=other_key_values)
 
     @staticmethod
     def _get_user_resource_relationship_clause(relation_type: UserResourceRel, id: str = None,
@@ -845,11 +1014,14 @@ class Neo4jProxy(BaseProxy):
                 user_matcher += ' {key: $user_key}'
 
         if relation_type == UserResourceRel.follow:
-            relation = f'(usr{user_matcher})-[rel:FOLLOW]->(resource{resource_matcher})'
+            relation = f'(resource{resource_matcher})-[r1:FOLLOWED_BY]->(usr{user_matcher})-[r2:FOLLOW]->' \
+                       f'(resource{resource_matcher})'
         elif relation_type == UserResourceRel.own:
-            relation = f'(usr{user_matcher})<-[rel:OWNER]-(resource{resource_matcher})'
+            relation = f'(resource{resource_matcher})-[r1:OWNER]->(usr{user_matcher})-[r2:OWNER_OF]->' \
+                       f'(resource{resource_matcher})'
         elif relation_type == UserResourceRel.read:
-            relation = f'(usr{user_matcher})-[rel:READ]->(resource{resource_matcher})'
+            relation = f'(resource{resource_matcher})-[r1:READ_BY]->(usr{user_matcher})-[r2:READ]->' \
+                       f'(resource{resource_matcher})'
         else:
             raise NotImplementedError(f'The relation type {relation_type} is not defined!')
         return relation
@@ -871,8 +1043,8 @@ class Neo4jProxy(BaseProxy):
 
         # FYI, to extract last_successful_execution, it searches for its execution ID which is always
         # _last_successful_execution
-        # https://github.com/lyft/amundsendatabuilder/blob/master/databuilder/models/dashboard/dashboard_execution.py#L18
-        # https://github.com/lyft/amundsendatabuilder/blob/master/databuilder/models/dashboard/dashboard_execution.py#L24
+        # https://github.com/amundsen-io/amundsendatabuilder/blob/master/databuilder/models/dashboard/dashboard_execution.py#L18
+        # https://github.com/amundsen-io/amundsendatabuilder/blob/master/databuilder/models/dashboard/dashboard_execution.py#L24
 
         query = textwrap.dedent(f"""
         MATCH {rel_clause}<-[:DASHBOARD]-(dg:Dashboardgroup)<-[:DASHBOARD_GROUP]-(clstr:Cluster)
@@ -1025,8 +1197,6 @@ class Neo4jProxy(BaseProxy):
                 tx.rollback()
             # propagate the exception back to api
             raise e
-        finally:
-            tx.close()
 
     @timer_with_counter
     def delete_resource_relation_by_user(self, *,
@@ -1049,9 +1219,9 @@ class Neo4jProxy(BaseProxy):
                                                                       )
 
         delete_query = textwrap.dedent("""
-        MATCH {rel_clause}
-        DELETE rel
-        """.format(rel_clause=rel_clause))
+                MATCH {rel_clause}
+                DELETE r1, r2
+                """.format(rel_clause=rel_clause))
 
         try:
             tx = self._driver.session().begin_transaction()
@@ -1062,8 +1232,6 @@ class Neo4jProxy(BaseProxy):
             if not tx.closed():
                 tx.rollback()
             raise e
-        finally:
-            tx.close()
 
     @timer_with_counter
     def get_dashboard(self,
@@ -1079,21 +1247,24 @@ class Neo4jProxy(BaseProxy):
         OPTIONAL MATCH (d)-[:LAST_UPDATED_AT]->(t:Timestamp)
         OPTIONAL MATCH (d)-[:OWNER]->(owner:User)
         WITH c, dg, d, description, last_exec, last_success_exec, t, collect(owner) as owners
-        OPTIONAL MATCH (d)-[:TAGGED_BY]->(tag:Tag)
-        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, collect(tag) as tags
+        OPTIONAL MATCH (d)-[:TAGGED_BY]->(tag:Tag{tag_type: $tag_normal_type})
+        OPTIONAL MATCH (d)-[:HAS_BADGE]->(badge:Badge)
+        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, collect(tag) as tags,
+        collect(badge) as badges
         OPTIONAL MATCH (d)-[read:READ_BY]->(:User)
-        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags,
+        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags, badges,
         sum(read.read_count) as recent_view_count
         OPTIONAL MATCH (d)-[:HAS_QUERY]->(query:Query)
-        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags,
-        recent_view_count, collect(query) as queries
+        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags, badges,
+        recent_view_count, collect({name: query.name, url: query.url, query_text: query.query_text}) as queries
         OPTIONAL MATCH (d)-[:HAS_QUERY]->(query:Query)-[:HAS_CHART]->(chart:Chart)
-        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags,
+        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags, badges,
         recent_view_count, queries, collect(chart) as charts
         OPTIONAL MATCH (d)-[:DASHBOARD_WITH_TABLE]->(table:Table)<-[:TABLE]-(schema:Schema)
         <-[:SCHEMA]-(cluster:Cluster)<-[:CLUSTER]-(db:Database)
         OPTIONAL MATCH (table)-[:DESCRIPTION]->(table_description:Description)
-        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags, recent_view_count, queries, charts,
+        WITH c, dg, d, description, last_exec, last_success_exec, t, owners, tags, badges,
+        recent_view_count, queries, charts,
         collect({name: table.name, schema: schema.name, cluster: cluster.name, database: db.name,
         description: table_description.description}) as tables
         RETURN
@@ -1112,43 +1283,54 @@ class Neo4jProxy(BaseProxy):
         toInteger(t.timestamp) as updated_timestamp,
         owners,
         tags,
+        badges,
         recent_view_count,
         queries,
         charts,
         tables;
         """
                                                      )
-        record = self._execute_cypher_query(statement=get_dashboard_detail_query,
-                                            param_dict={'query_key': id}).single()
+        dashboard_record = self._execute_cypher_query(statement=get_dashboard_detail_query,
+                                                      param_dict={'query_key': id,
+                                                                  'tag_normal_type': 'default'}).single()
 
-        if not record:
+        if not dashboard_record:
             raise NotFoundException('No dashboard exist with URI: {}'.format(id))
 
-        owners = [self._build_user_from_record(record=owner) for owner in record['owners']]
-        tags = [Tag(tag_type=tag['tag_type'], tag_name=tag['key']) for tag in record['tags']]
-        chart_names = [chart['name'] for chart in record['charts'] if 'name' in chart and chart['name']]
-        query_names = [query['name'] for query in record['queries'] if 'name' in query and query['name']]
-        tables = [PopularTable(**table) for table in record['tables'] if 'name' in table and table['name']]
+        owners = [self._build_user_from_record(record=owner) for owner in dashboard_record['owners']]
+        tags = [Tag(tag_type=tag['tag_type'], tag_name=tag['key']) for tag in dashboard_record['tags']]
 
-        return DashboardDetailEntity(uri=record['uri'],
-                                     cluster=record['cluster_name'],
-                                     url=record['url'],
-                                     name=record['name'],
-                                     product=record['product'],
-                                     created_timestamp=record['created_timestamp'],
-                                     description=self._safe_get(record, 'description'),
-                                     group_name=self._safe_get(record, 'group_name'),
-                                     group_url=self._safe_get(record, 'group_url'),
-                                     last_successful_run_timestamp=self._safe_get(record,
+        badges = [TableBadge(badge_name=badge['key'],
+                             category=badge['category']) for badge in dashboard_record['badges']]
+
+        chart_names = [chart['name'] for chart in dashboard_record['charts'] if 'name' in chart and chart['name']]
+        # TODO Deprecate query_names in favor of queries after several releases from v2.5.0
+        query_names = [query['name'] for query in dashboard_record['queries'] if 'name' in query and query['name']]
+        queries = [DashboardQueryEntity(**query) for query in dashboard_record['queries']
+                   if query.get('name') or query.get('url') or query.get('text')]
+        tables = [PopularTable(**table) for table in dashboard_record['tables'] if 'name' in table and table['name']]
+
+        return DashboardDetailEntity(uri=dashboard_record['uri'],
+                                     cluster=dashboard_record['cluster_name'],
+                                     url=dashboard_record['url'],
+                                     name=dashboard_record['name'],
+                                     product=dashboard_record['product'],
+                                     created_timestamp=dashboard_record['created_timestamp'],
+                                     description=self._safe_get(dashboard_record, 'description'),
+                                     group_name=self._safe_get(dashboard_record, 'group_name'),
+                                     group_url=self._safe_get(dashboard_record, 'group_url'),
+                                     last_successful_run_timestamp=self._safe_get(dashboard_record,
                                                                                   'last_successful_run_timestamp'),
-                                     last_run_timestamp=self._safe_get(record, 'last_run_timestamp'),
-                                     last_run_state=self._safe_get(record, 'last_run_state'),
-                                     updated_timestamp=self._safe_get(record, 'updated_timestamp'),
+                                     last_run_timestamp=self._safe_get(dashboard_record, 'last_run_timestamp'),
+                                     last_run_state=self._safe_get(dashboard_record, 'last_run_state'),
+                                     updated_timestamp=self._safe_get(dashboard_record, 'updated_timestamp'),
                                      owners=owners,
                                      tags=tags,
-                                     recent_view_count=record['recent_view_count'],
+                                     badges=badges,
+                                     recent_view_count=dashboard_record['recent_view_count'],
                                      chart_names=chart_names,
                                      query_names=query_names,
+                                     queries=queries,
                                      tables=tables
                                      )
 
